function [fitness_val, total_connections] = fitness_mult(A, G, rev_ord, MEX, TempNoiseFlag, NoiseIntensity)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Written by Tamar Friedlander, IST Austria
% June 2014
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This function receives as input the population of multilayered networks
% ('A') and the goal matrix 'G' and then evalues fitness by calculating the
% difference between the product the of the A matrices and the goal: F =
% -||A{L}*A{L-1}*....*A{1} - G||.


% A is a cell in which A{i} represents connection intensities between i-th
% and i+1-th layers of nodes in the model.
% calculate fitness of the matrix product A{1}*A{2}*...*A{L}, where the goal system is G.
% This function makes use of the mex-function 'mtimesx' for fast
% multiplication of matrices. It can be downloaded from Matlab website: 
% http://www.mathworks.com/matlabcentral/fileexchange/25977-mtimesx-fast-matrix-multiply-with-multi-dimensional-support

% trace(A*A') equals the sum of the square values of all matrix elements.
% 7/2/2012: trace is elegant but much more time consuming (takes 10^-3 sec
% instaed of 10^-5 for a simple double sum). Since this function consumes about half run-time (!) 'trace' was thus replaced by a
% double sum.

% 16/7/2013: I reversed the order of matrix multiplication to be consistent
% with other works. Until now it was A{1}*A{2}*A{3}*...A{L} so that the input
% was layer L and output was layer 1.
% From now on it is A{L}*A{L-1}*....*A{1} so that input is layer a and
% output is layer L. If the flag rev_ord =0 assume the old order. Other
% wise use the new order.

% 28/7/2013: 
% for speed-up:
% since G and PopSize do not change throughout the simulation claculate these once in the beginning and use as a parameter for the function instead of calling 'repmat' at each generation.
% repmat(G, [1 1 PopSize])
% 'G' was replaced by a replicated version 'G_rep' = repmat(G,
% [1 1 PopSize]); This save applying 'repmat' again every generation.

% % 26/8/2013: changed connection cost from sum of squares (concave function) to sum of sqrt (convex).

% 31/12/2013: change in the definition of fitness for non-zero epsilon.
% From now on 'fitness_val' represents only distance from the goal. The
% 'total_connections' variable is also computed and weighted fitness can be
% computed outside the function.

% 31/3/2014: the loop calculating 'total-connections' is disabled because
% it is very time consuming. Find a more efficient way to write it. At the
% meantime the output value of this variable is zero. 

% 24/4/2014: Add a temporal noise option: Fitness is now F=-||(A_L +
% e_L)*(A_L-1 + e_L-1)*...*(A_1 + e_1) - (G + e_G)||


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

L = size(A,2);
if L < 2,
    error('A has only a single layer. This routine is inappropriate.');
end;
PopSize = size(A{1},3);
if TempNoiseFlag,
    M = cell(1,L);
    Noise = cellfun(@(x) randn(size(A{1})) * NoiseIntensity, M, 'un',0);
    A = cellfun(@plus, A, Noise,'UniformOutput', false); 
end;

if MEX
    if rev_ord == 0,
        C = A{1};
        for ii = 2:L,
            C = mtimesx(C,A{ii});
        end;
    else % if rev_ord == 1
        C = A{L};
        for ii = L-1:-1:1,
            C = mtimesx(C,A{ii});
        end;
    end;
else % MEX == 0
    C = A{L};
    for nn = 1: PopSize,
        for ii = L-1:-1:1
            Temp = C(:,:,nn) * A{ii}(:,:,nn);
            C(:,:,nn) = Temp; 
        end;
    end;
end;
    

fitness_val = - squeeze(sum(sum   ( (C - G).^2)));

total_connections = zeros(1,PopSize);
% for ii = 1:L
%     total_connections = total_connections + squeeze(sum(sum(sqrt(abs(A{ii})) )))'; % 26/8/2013: changed from sum of squares (concave function) to sum of sqrt (convex).
%     %fitness_val = fitness_val - epsilon * squeeze(sum(sum(A{ii}.^2)));
% end;


