function [] = ga_multiple_layers(IntSeed, GA, FLAGS, G, fname)

%********************************************************
% multi-layer network evolutionary simulations
% 
% written by Tamar Friedlander, June 2014
%********************************************************

% 7/7/2013
% Extension of the previous model to account for a general number of layers
% (rather than 3).
% based on the version submitted to Dryad.
% Removed the option for alternating goals.

% 16/7/2013: I reversed the order of matrix multiplication to be consistent
% with other works. Until now it was A{1}*A{2}*A{3}*...A{L} so that the input
% was layer L and output was layer 1.
% From now on it is A{L}*A{L-1}*....*A{1} so that input is layer 1 and
% output is layer L. If the flag rev_ord =0 assume the old order. Other
% wise use the new order.

% 26/3/2014: Biased additive mutations option was added,
% GA.mutation.MutType == 6. The mutated value is multiplied by a constant
% (generally < 1 to get a bias to decrease, if it equals 1 this is
% equivalent to sum mutation) and then additive noise is added. 

%  31/3/2014: GA.goal.GoalType = 2 was added. Temporally noisy goal. The
%  goal noise component is updated every generation. Note that with this
%  option the variable 'G' that is the input is the clean goal. Noise is
%  added in this function every generation. 

% 4/4/2014: if the goal is temporally noisy (GA.goal.GoalType = 2),
% 'mean_fitness' is calculated with respect to the clean goal (temporal
% average over the noisy goals actually).

 % 24/4/2014: Temporal noise in fitness calculation is enabled
 % (additional arguments to 'fitness_mult').
 % Fitness is now F = -||(A_L + e_L)*(A_L-1 + e_L-1)*...*(A_1 + e_1) - (G + e_G)||

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% function definitions:
my_cat = @(A,B) cat(3,A,B);
            
rand ('state', IntSeed);  % Initialize RAND to a preset state, so that runs could be repeated.
randn('state', IntSeed);  % Initialize RAND to a preset state, so that runs could be repeated.

StopCon = 0;
mean_fitness = zeros(1,GA.general.MaxGen);
best_fitness = zeros(1,GA.general.MaxGen);
mean_connections = zeros(1,GA.general.MaxGen);

if  GA.goal.GoalType > 0
    if size(G) ~= GA.goal.MatDim * ones(1,2) %| size(G2) ~= GA.goal.MatDim * ones(1,2),
        error('G dimensions don''t match A and B dimensions!!!');
    end;
end;
   
%%%%%%%%%%%%%%%    
% initialize  %
%%%%%%%%%%%%%%%
L = GA.general.MatNum;
A_Pop = cell(1,L);  
MutatedGenes = cell(1,L); UnMutatedGenes = cell(1,L); A_MutPop = cell(1,L);
for ii = 1:L,
    A_Pop{ii} = GA.goal.LB + (GA.goal.UB - GA.goal.LB) * rand(GA.goal.MatDim,  GA.goal.MatDim,  GA.general.PopSize); % population represented as a 3D matrix. Entails a loop in fitness calculation.
end;
  
if GA.init.FileFLAG % load initial population from a file
    addpath 'directory where file is kept'
    load('FileName.mat'); % found in 'D:\My Documents\modularity\simulations\ver_6', and represents a FG run of 3000 generations at D=4, RandSeed=1.
    cd 'simulation directory'
end;
% create a replicated version of G for fitness calculations, rather than
% replicating it each generation again.
G_rep = repmat(G, [1 1 2*GA.general.PopSize]);
 if GA.goal.GoalType == 2,
     clean_G = G;
     clean_G_rep = repmat(G, [1 1 GA.general.PopSize]);
 end;
gen = 1;     

%%%%%%%%%%%%%%%%%%%%
% generation loop  %
%%%%%%%%%%%%%%%%%%%%
   while (StopCon == 0) && (gen <= GA.general.MaxGen)  , % stopping condition is not met and maximal number of generations not exceeded.
       if GA.goal.GoalType == 2 % added 31/3/2014: '2' - Temporally noisy goal. The goal noise component is updated every generation. 
            G = clean_G + GA.goal.TempNoise * randn(GA.goal.MatDim);
            G_rep = repmat(G, [1 1 2*GA.general.PopSize]);
       end;

       %%%%%%%%%%%%%%%%%%%%%%%%%%
       %         Mutation       %
       %%%%%%%%%%%%%%%%%%%%%%%%%%

       for ii=1:L,
           MutatedGenes{ii} = rand(size(A_Pop{ii})) < GA.mutation.MutRate_A;  % choose which genes to mutate
       end;
       UnMutatedGenes = cellfun(@not, MutatedGenes, 'UniformOutput', false);
       
       if GA.mutation.MutType ~= 3 
             MM = randn ([size(A_Pop{1}) L]) * GA.mutation.MutSize_A;
             MutVals = (squeeze(num2cell(MM, [1 2 3])))';
       end;  % if GA.mutation.MutType == 1 || GA.mutation.MutType == 2,
       for ii = 1 : L,
           if GA.mutation.MutType == 1, % additive mutations
                A_MutPop{ii} = A_Pop{ii} + MutVals{ii}.*MutatedGenes{ii};  % additive mutations. No upper and lower bounds.
           elseif GA.mutation.MutType == 2, % 4/9/2012: change. probability to change sign was added. To return to the previous setting, set the parameter GA.mutation.sgn_switch_prob to zero.
                A_MutPop{ii} = A_Pop{ii}.*(UnMutatedGenes{ii} + MutatedGenes{ii}.* (GA.mutation.MeanMut + MutVals{ii}).* sign (rand(size(A_Pop{ii})) - GA.mutation.sgn_switch_prob)) ;  % multiplicative mutations. No upper and lower bounds. 
                % disable these lines for speed
                MutVals2 = randn (size(A_Pop{ii})) * GA.mutation.MutSize2_A;
                A_MutPop{ii} = A_MutPop{ii} + MutVals2 .* MutatedGenes{ii};  % 20/12/2011: small additive mutation occurs only in the same genes where there was a multiplicative mutation. In earlier versions this additive mutation was to all genes. 
           elseif GA.mutation.MutType == 3, %%% Log-normal distribution of mutations.
                A_MutPop = A_Pop.* (1 + (MutVals - 1).*MutatedGenes); 
                MutVals2 = randn (size(A_Pop)) * GA.mutation.MutSize2_A;
                A_MutPop = A_MutPop + MutVals2;
           elseif GA.mutation.MutType == 4, %% additive non-negative - absorbing boundary. added 28/8/2012.
                %A_MutPop = max(0, A_Pop + MutVals.*MutatedGenes);
                A_MutPop = A_Pop.*UnMutatedGenes + MutatedGenes.*(max(0, A_Pop + MutVals.*MutatedGenes)); % corrected 9/9/2012.
           elseif GA.mutation.MutType == 5, %% additive non-negative - reflecting boundary. added 4/9/2012.
                % actually takes absolute value of all terms, so if we
                % start with negative values, they will not survive.
                A_MutPop = A_Pop.*UnMutatedGenes + MutatedGenes.*(abs(A_Pop + MutVals.*MutatedGenes)); % corrected 9/9/2012
           elseif GA.mutation.MutType == 6, % biased additive mutations; If GA.mutation.MeanMut=1, this should coincide with sum-mutations (Type 1); added 26/03/2014
                A_MutPop{ii} = A_Pop{ii}.*(MutatedGenes{ii}*GA.mutation.MeanMut + ~MutatedGenes{ii}) + MutVals{ii}.*MutatedGenes{ii};  % additive mutations. No upper and lower bounds.
           end;
       end;
       % keep 2 copies: the origianl (unmutated) and the mutated one, and
       % then select N times from these 2N individuals.
       %my_cat = @(A,B) cat(3,A,B);
       A_temp = cellfun(my_cat,A_Pop,A_MutPop,'UniformOutput', false);

       %%%%%%%%%%%%%%%%%%%%%%%%
       % Fitness Evaluation   %
       %%%%%%%%%%%%%%%%%%%%%%%%
       [fval,connections] = fitness_mult (A_temp, G_rep, GA.general.rev_ord, FLAGS.MEX, GA.Noise.TempNoiseFLAG, GA.Noise.Intensity); % evaluate population fitness
       
       % changed 31/12/2013: 'mean_fitness' record refers only to distance
       % from the goal and not to connection costs. 
       % 4/4/2014: If the goal has temporal noise the recorded
       % 'mean_fitness' will refer to the clean goal (temporally average
       % goal and not to the different instances that include noise. So
       % actually this recorded value is different than the one that is
       % used for the selection. 
       % 24/4/2014: Temporal noise in fitness calculation is enabled
       % (additional arguments to 'fitness_mult').
       if GA.goal.GoalType == 2, % temporally changing noise (on top of the goal) - calculate fitness again, now with the clean goal.
           [fval_clean,connections] = fitness_mult (A_Pop, clean_G_rep, GA.general.rev_ord, FLAGS.MEX, 0); % evaluate population fitness. Here no noise is taken into consideration. 
           mean_fitness(gen) = mean(fval_clean); % changed 11/7/2013: calculate mean only over the parent fitness. Otherwise few bad mutatnts that will not be selected can significantly bias the population mean.
           mean_connections(gen) = mean(connections(1:GA.general.PopSize)); 
           best_fitness(gen) = max(fval_clean);
       else
           mean_fitness(gen) = mean(fval(1:GA.general.PopSize)); % changed 11/7/2013: calculate mean only over the parent fitness. Otherwise few bad mutatnts that will not be selected can significantly bias the population mean.
           mean_connections(gen) = mean(connections(1:GA.general.PopSize)); 
           best_fitness(gen) = max(fval);
       end; % % temporally changing noise (on top of the goal)
       
       % changed 31/12/2013: weighted fitness (including connection costs)
       % is calculated here and not in the 'fitness_mult' function.
       if GA.general.epsilon~=0
           fval = fval - GA.general.epsilon * connections';
       end;
       % 16/7/2013: The parameter 'GA.general.rev_ord' which can reverse
       % the order of matrix multiplication was added. 
       % 18/8/2013: epsilon was added as an additional input parameter that
       % represents selection to minimize connection costs. 

       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       % Fitness Scaling               %
       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       if strcmp (GA.selection.fitscale, 'Exponential')
           scaled_fval = exp(GA.selection.beta*fval) / sum(exp(GA.selection.beta*fval)); % quicker, but doesn't work in high dimensions.
           while sum(isnan(scaled_fval))~=0,
               scaled_fval = scale_f (fval, GA.selection.beta); % slower, but more robust to numerical precision problems.
           end;
       elseif strcmp (GA.selection.fitscale, 'None')
           scaled_fval = fval;
       end; % if GA.selection.fitscale == 'Exponential'
       

       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       %       Selection         %
       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       
       % The population needs to be heterogeneous to some extenet for this
       % algorithm to work. If the population is completely homogeneous,
       % then it is kept unchanged, hoping that mutations will create some
       % diversity in the future.
       
      if strcmp (GA.selection.method, 'Tournament')
             tournament = randi (2 * GA.general.PopSize, [GA.selection.tournament_s, GA.general.PopSize]); % choose 'PopSize' sets of 's' individuals each with uniform distribution.
             [~, ord] = sort(scaled_fval,'descend'); % first ones are the best.
             chosen_ind = min(tournament); % choose the fittest one from each set of 's' individuals.
            for ii = 1 : L, % can this be written without a loop???
                A_temp{ii} = A_temp{ii}(:,:,ord);
                A_Pop{ii} = A_temp{ii}(:,:,chosen_ind);
            end;
       elseif strcmp (GA.selection.method, 'Roulette') % roulette wheel selection
           if std(scaled_fval)~=0, % heterogeneous fitness population
               J = randsample(2*GA.general.PopSize, GA.general.PopSize, true, scaled_fval)'; 
               for ii = 1 : L,
                   A_Pop{ii} = A_temp{ii}(:,:,J); % is there a way to write this without the loop???
               end;
           end;
       end;  %       if GA.selection.method == ...

       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       %  evaluate simulation stop condition   %
       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if strcmp(GA.general.StopCon,'FitnessOnly')  % 20/3/2012: if additive mutations are used, the stopping condition cannot require a number of zeros, but only a fitness criterion.
       if abs(mean_fitness(gen)) < GA.general.StopThreshold % 24/4/2012: defined the parameter 'GA.general.StopThreshold'.
           StopCon = 1;
       end;
   elseif strcmp(GA.general.StopCon,'None') % Stop only by number of generations
       StopCon = 0;
   end;
   gen = gen + 1;
   end; % while StopCon == 0, % stopping condition is not met
mean_fitness = mean_fitness(1:gen - 1);
best_fitness = best_fitness(1:gen - 1);


if FLAGS.save_FLAG,
   s = rng; % save current random seed value.
   cd (GA.general.save_path);
   if exist ('G','var')
           rand_ind = randi(GA.general.PopSize, 1);
           A = cell(1,L);
           for ii=1:L, % save only one representative
               A{ii} = A_Pop{ii}(:,:,rand_ind);
           end;
           if FLAGS.save.full
               save (fname,'mean_fitness','A','GA','gen','G','s'); %,'mean_connections');
           else
               mean_fitness_end = mean_fitness(end);
               mean_connections_end = mean_connections(end);
               save (fname,'A','GA','gen','G','s','gen','mean_fitness_end');  % 24/2/2014: save also the number of generations performed in the simulation. 
           end;
   end;

end; % if FLAGS.save_FLAG,